unit uSXB2XB;

{
This code is part of the TiCodEd Application, designed to develop for the
TI-99/4a Homecomputer in a modern, structured way.

BSD License
Copyright (c) 2020, Stefan Bauch, TiCodEd <at> lizardware <dot> de


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type TLogProc = Procedure(l:Integer; txt:String);

     TLine = Record
       nr : integer;
       lbl: String[32];
       stmt : String;
     end;

     TFType = (f_label,f_linenum,f_keyword, f_string, f_cond);

     lkind = (lwhile,lrepeat);
     TLoop = Record
       kind : lkind;
       num  : Integer;
     end;

Var debug : Boolean = true;
    prg   : Array[1..512] of TLine;
    maxl  : Integer;

Function SXBtoXB(sxbin,xbout: String;linestart,lineinc:Integer;WriteLogProc:TLogProc;Options:String):Integer;


implementation

uses ctype;


Type  TLbl = Record
               txt : String[32];
               line: Integer;
             end;


VAR  WriteLog : TLogProc;
     sxb, xb : Textfile;
     lstack : Array[0..31] of TLoop;
     lstptr : Integer = 0;

Procedure dbg(l:Integer; txt:String);
begin
  writelog(l,txt);
end;


function isstop(c:char):boolean;
begin
  if c in [' ',':','(',','] then isstop := true
                            else isstop := false;
end;

procedure lpush(lk : lkind; nm : Integer);
begin
  lstack[lstptr].kind := lk;
  lstack[lstptr].num  := nm;
  inc(lstptr);
end;

procedure lpop(VAR lk : lkind; VAR nm : Integer);
begin
  dec(lstptr);
  if lstptr>=0 then begin
    lk := lstack[lstptr].kind;
    nm := lstack[lstptr].num;
  end else dbg(1,'ERROR: REPEAT/WHILE nesting error.');
end;


function cmdend(s:string):Integer;
var inquote:boolean;
    i:Integer;
begin
  inquote := false;
  i := 1;
  s:=s+' ';
  if uppercase(leftstr(s,5))='DATA ' then cmdend:=length(s)-1 else begin
     for i:=1 to length(s) do begin
       if s[i]='"' then inquote := not inquote;
       if inquote then s[i]:='*'
     end;
     if pos('::',s)= 0 then cmdend:=length(s)-1
                       else cmdend:=pos('::',s)-1
  end;
end;


function PosUnquoted(sub,s:string):Integer;
var inquote:boolean;
    i:Integer;
begin
  inquote := false;
  i := 1;
  sub := uppercase(sub);
  s:=uppercase(s)+' ';
  if leftstr(s,5)='DATA ' then PosUnquoted:=0 else begin
       for i:=1 to length(s) do begin
         if s[i]='"' then inquote := not inquote;
         if inquote then s[i]:=#255;
       end;
       PosUnquoted := pos(sub,s);
  end;
end;


Procedure buildlist(lstart,linc: Integer);
var s,t: string;
    i,j,l : Integer;
    nm, loopcount,maxlbl : Integer;
    lk : lkind;
    sorted : boolean;
    labtab: Array[0..99] of Tlbl;
    templab : Tlbl;

begin
  l := 1;
  loopcount := 0;
  while not eof(sxb) do begin
    readln(sxb,s);
    s := trim(s) + ' '; // blank to avoid checking length in s[i+1]
    if s[1]='#' then if debug
       then s := trim(copy(s,2,254))
       else s := '// ' + s;

    // init line-record
    prg[l].nr   := 0;
    prg[l].lbl  := '';
    prg[l].stmt := '';

    i:=1;
    // Line-Number ?
    while (i<length(s)) and isdigit(s[i]) do inc(i);
    if (i>1) and (s[i]=#32) then begin
       t := leftstr(s,i-1);
       dbg(4,'Line-nr: '+t);
       prg[l].nr   := Strtoint(t);
       prg[l].stmt := trim(copy(s,i+1,255)) + ' ';
    end else begin // Label ?
      i:=1;
      while (i<length(s)) and isalnum(s[i]) do inc(i);
      if (i>1) and (s[i]=':') then begin
         t := uppercase(leftstr(s,i-1));
         dbg(4,'Label: '+t);
         prg[l].lbl  := t;
         prg[l].stmt := trim(copy(s,i+1,255)) + ' ';
      end else prg[l].stmt:=s;
    end;
    dbg(4,'Statement: '+prg[l].stmt);
    if (Uppercase(trim(prg[l].stmt))='ENDWHILE') or
       (Uppercase(trim(prg[l].stmt))='WEND') then begin
       // Create additional line for ENDWHILE-Label
       dbg(5,'Create addtional Line for ENDWHILE-Label');
       inc(l);
       prg[l].nr   := 0;
       prg[l].lbl  := '<ENDWHILE>';
       prg[l].stmt := '';
    end;
    inc(l);
  end; // while not EOF ...
  maxl := l - 1;
  // Replaces REPEAT ... UNTIL
  for l:=1 to maxl do begin
    if Uppercase(leftstr(prg[l].stmt,7))='REPEAT ' then begin
      prg[l].stmt := trim(copy(prg[l].stmt,7,255))+' ';
      inc(loopcount);
      lpush(lrepeat,loopcount);
      prg[l].lbl  := 'REPEAT'+format('%.3d',[loopcount]);
      dbg(3,'REPEAT: '+prg[l].lbl+':')
    end else if Uppercase(leftstr(prg[l].stmt,6))='UNTIL ' then begin
      t := copy(prg[l].stmt,7,cmdend(prg[l].stmt)-7);
      lpop(lk,nm);
      if lk=lwhile then dbg(0,'ERROR in Nesting: ENDWHILE expected in line '+inttostr(i));
      dbg(3,'UNTIL ['+t+']');
      prg[l].stmt := 'IF NOT ('+t+') THEN GOTO REPEAT'+format('%.3d',[nm])+copy(prg[l].stmt,cmdend(prg[l].stmt),255);
      dbg(4,prg[l].stmt);
    end;
  end;

  // Replaces WHILE ... ENDWHILE
  for l:=1 to maxl do begin
    if Uppercase(leftstr(prg[l].stmt,6))='WHILE ' then begin
      inc(loopcount);
      lpush(lwhile,loopcount);
      t := copy(prg[l].stmt,7,cmdend(prg[l].stmt)-7);
      dbg(3,'WHILE ['+t+']');
      prg[l].stmt := 'IF NOT ('+t+') THEN GOTO ENDWHILE'+format('%.3d',[loopcount])+copy(prg[l].stmt,cmdend(prg[l].stmt),255);
      prg[l].lbl  := 'WHILE'+format('%.3d',[loopcount]);
      dbg(4,prg[l].stmt);
    end else if (Uppercase(leftstr(prg[l].stmt,9))='ENDWHILE ') or
                (Uppercase(leftstr(prg[l].stmt,5))='WEND ') then begin
       lpop(lk,nm);
       if lk=lrepeat then dbg(0,'ERROR in Nesting: UNTIL expected in line '+inttostr(i));
       prg[l].stmt := 'GOTO WHILE'+format('%.3d',[nm]);
       // Use additional line for ENDWHILE-Label
       dbg(4,'Check: Line for ENDWHILE? '+ prg[l+1].lbl);
       prg[l+1].lbl  := 'ENDWHILE'+format('%.3d',[nm]);
    end;
  end;

  // Remove "//" Comments
  for l:=1 to maxl do begin
     j := PosUnquoted('//',prg[l].stmt);
     if j>0 then begin
        dbg(4,'Remove comment '+copy(prg[l].stmt,j,255));
        prg[l].stmt := copy(prg[l].stmt,1,j-1);
     end;
  end;


  // Assign Line-Numbers to all lines, -1 means empty line to be skipped
  for l:=1 to maxl do with prg[l] do begin
     if (trim(stmt) = '') and (nr=0) then nr := -1
    else if nr=0 then begin
       nr := lstart;
       inc(lstart,linc);
    end else if nr>lstart then begin
       dbg(4,'Overwriting Line-Number '+inttostr(lstart)+' with '+inttostr(nr));
       lstart := nr;
       if trim(stmt)='' then nr:=-1;
    end else if nr<lstart then dbg(0,'ERROR in line-numbering! '+inttostr(lstart)+' larger than '+inttostr(nr))
  end;

  // Build table of Labels with Line-Numbers
  dbg(3,'');
  dbg(3,'Line Number Table');
  maxlbl := 0;
  for l:=1 to maxl do with prg[l] do
     if (lbl<>'') then begin
        j:=l;
        // assign the next valid line-number
        while prg[j].nr=-1 do inc(j);
        labtab[maxlbl].txt  := lbl;
        labtab[maxlbl].line := prg[j].nr;
        inc(maxlbl);
     end;
  dec(maxlbl);
  // Sort Labels by descending length
  repeat
     sorted := true;
     for j:=0 to maxlbl-1 do
       if length(labtab[j].txt)<length(labtab[j+1].txt) then begin
         templab     := labtab[j];
         labtab[j]   := labtab[j+1];
         labtab[j+1] := templab;
         sorted      := false;
       end;
  until sorted;

  for j:=0 to maxlbl do dbg(3,'   '+        labtab[j].txt+' -> '+inttostr(labtab[j].line));

  // Replace Labels with Line-Numbers
  for l:=1 to maxl do
    if prg[l].nr>0 then begin
      for j:=0 to maxlbl do begin
         i := PosUnquoted(' '+labtab[j].txt, prg[l].stmt);
         while i > 0  do begin
            dbg(4,'Label replace');
            dbg(4,'     '+ prg[l].stmt);
            prg[l].stmt := copy(prg[l].stmt,1,i)
                         + inttostr(labtab[j].line)
                         + copy(prg[l].stmt,i+length(labtab[j].txt)+1,255);
            i := PosUnquoted(' '+labtab[j].txt, prg[l].stmt);
            dbg(4,'  -> ' + prg[l].stmt);
         end;
      end;
    end;

  dbg(4,'');
  dbg(4,'Standard Extended Basic:');
  dbg(4,'');
  // write XB File
  for i:=1 to maxl do with prg[i] do if nr>0 then begin
    dbg(4,inttostr(nr)+' '+stmt);
    writeln(xb,nr,' ',trim(stmt));
  end;
end;



Function SXBtoXB(sxbin,xbout: String;linestart,lineinc:Integer;WriteLogProc:TLogProc;Options:String):Integer;
begin
  WriteLog := WriteLogProc;
  assign(sxb,sxbin);
  reset(sxb);
  assign(xb,xbout);
  rewrite(xb);

  if pos('-DEBUG', UpperCase(Options)) > 0 then debug := true
                                           else debug := false;

  dbg(2,'starting to convert SXB to XB ....');
  if debug then dbg(3,'Debug is active: Lines starting # will be included');

  buildlist(linestart,lineinc);

  system.close(sxb);
  system.close(xb);

  sxbtoXB := 0;
end;

end.

