unit uMain;
{
This code is part of the TiCodEd Application, designed to develop for the
TI-99/4a Homecomputer in a modern, structured way.

BSD License
Copyright (c) 2020, Stefan Bauch, TiCodEd <at> lizardware <dot> de

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Menus, ExtCtrls,
  ComCtrls, StdCtrls, SynHighlighterVB, SynEdit, IniFiles, LCLType,
  ColorBox, MaskEdit, tidbit, uSXB2XB, uTokenize;

type

  { TForm1 }

  TForm1 = class(TForm)
    cbtkn: TCheckBox;
    cbDebug: TCheckBox;
    cbExBas: TCheckBox;
    cbMrg: TCheckBox;
    ColorBoxFG: TColorBox;
    ColorBoxBG: TColorBox;
    EdFileSXB: TEdit;
    EdFilePrj: TEdit;
    EdFileXBtxt: TEdit;
    EdFileXB: TEdit;
    EdFileXBM: TEdit;
    GroupBox4: TGroupBox;
    hexstr: TEdit;
    GroupBox1: TGroupBox;
    GroupBox2: TGroupBox;
    GroupBox3: TGroupBox;
    Label1: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    Log: TMemo;
    MaskEdit1: TMaskEdit;
    MaskEdit2: TMaskEdit;
    MenuUndo: TMenuItem;
    MenuNew: TMenuItem;
    MnReplace: TMenuItem;
    MnEdit: TMenuItem;
    MenuItemLogSave: TMenuItem;
    MenuItemLogClear: TMenuItem;
    PopupMenuLog: TPopupMenu;
    RadioGroup1: TRadioGroup;
    ReplaceDialog1: TReplaceDialog;
    CharPrev: TShape;
    StatusBar1: TStatusBar;
    TabSheet5: TTabSheet;
    TrackBarVerbose: TTrackBar;
    xb: TMemo;
    MainMenu1: TMainMenu;
    MenuSXBtoXB: TMenuItem;
    MenuFile: TMenuItem;
    MenuExit: TMenuItem;
    MenuBuild: TMenuItem;
    MenuTokenize: TMenuItem;
    MenuItem1: TMenuItem;
    MenuTIdBit: TMenuItem;
    MenuOpen: TMenuItem;
    MenuSave: TMenuItem;
    OpenDialog1: TOpenDialog;
    PageControl1: TPageControl;
    SynEdit1: TSynEdit;
    SynVBSyn1: TSynVBSyn;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    TabSheet4: TTabSheet;
    procedure CharPrevMouseEnter(Sender: TObject);
    procedure ColorBoxBGChange(Sender: TObject);
    procedure ColorBoxFGChange(Sender: TObject);
    procedure ColorBoxFGGetColors(Sender: TCustomColorBox; Items: TStrings);
    procedure FormCreate(Sender: TObject);
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
    procedure GroupBox1Resize(Sender: TObject);
    procedure hexstrChange(Sender: TObject);
    procedure hexstrEnter(Sender: TObject);
    procedure hexstrExit(Sender: TObject);
    procedure MenuBuildClick(Sender: TObject);
    procedure MenuExitClick(Sender: TObject);
    procedure MenuItem1Click(Sender: TObject);
    procedure MenuItemLogSaveClick(Sender: TObject);
    procedure MenuItemLogClearClick(Sender: TObject);
    procedure MenuNewClick(Sender: TObject);
    procedure MenuSXBtoXBClick(Sender: TObject);
    procedure MenuOpenClick(Sender: TObject);
    procedure MenuSaveClick(Sender: TObject);
    procedure MenuTIdBitClick(Sender: TObject);
    procedure MenuTokenizeClick(Sender: TObject);
    procedure MenuUndoClick(Sender: TObject);
    procedure MnReplaceClick(Sender: TObject);
    procedure PageControl1Change(Sender: TObject);
    procedure RadioGroup1Click(Sender: TObject);
    procedure CharPrevMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ReplaceDialog1Find(Sender: TObject);
    procedure ReplaceDialog1Replace(Sender: TObject);
    procedure ReplaceDialog1Show(Sender: TObject);
    procedure TrackBarVerboseChange(Sender: TObject);
    procedure xbDblClick(Sender: TObject);
  private

  public

  end;


var
  Form1: TForm1;
  sxbfile : String;
  prj : TIniFile;
  CharBit : Array[0..255] of TShape;
  clfg, clbg : TColor;
  hexmanmode : boolean;
  errcount, wrncount: Integer;


implementation

{$R *.lfm}

{ TForm1 }

uses math,ctype,SynEditTypes, LCLIntf;

Procedure WriteLogLine(l:integer; s:String);
begin
  if l<=Form1.TrackBarVerbose.Position then Form1.log.Lines.Append(s);
  if l=0 then inc(errcount);
  if l=1 then inc(wrncount);
end;


Function LoadFile(sxb : String):boolean;
begin
  loadfile := false;
  if fileexists(sxb) then with form1 do begin
    Form1.SynEdit1.Lines.LoadFromFile(sxb);
    Form1.Caption:='TI Code Editor - '+sxbfile;
    LoadFile := true;
    sxbfile  := sxb;
    EdFileSXB.Text:=sxb;
    EdFilePrj.Text := ChangeFileExt(sxb,'.xbp');
    prj := TiniFile.Create(EdFilePrj.Text);
    EdFileXBtxt.Text := prj.ReadString('Files','XBtxt',ChangeFileExt(sxb,'.xb'));
    EdFileXB.Text    := prj.ReadString('Files','XB',ChangeFileExt(sxb,''));
    EdFileXBM.Text   := prj.ReadString('Files','Merge',ChangeFileExt(sxb,'')+'-M');
    cbtkn.checked    := prj.ReadBool('Build','Token',True);
    cbMrg.checked    := prj.ReadBool('Build','Merge',True);
    MaskEdit1.Text   := prj.ReadString('Param','LineStart','100');
    MaskEdit2.Text   := prj.ReadString('Param','LineInc','10');
    cbDebug.checked  := prj.ReadBool('Param','Debug',True);
    TrackBarVerbose.Position := prj.ReadInteger('Param','Verbose',3);

    log.Lines.Clear;
    xb.Lines.Clear;
    ErrCount:=0; WrnCount := 0;
    WriteLogLine(-1,'File '+sxbfile+' loaded.');
  end else Application.MessageBox(pChar('File '+sxb+' not found.'), 'Error',0 );
end;

Procedure SaveFile;
begin
  with Form1 do begin
    if sxbfile = '' then begin
      Opendialog1.Execute;
      sxbfile := OpenDialog1.FileName;
      Form1.Caption:='TI Code Editor - '+sxbfile;
      EdFileSXB.Text:=sxbfile;
      EdFilePrj.Text := ChangeFileExt(sxbfile,'.xbp');
      prj := TiniFile.Create(EdFilePrj.Text);
      EdFileXBtxt.Text := prj.ReadString('Files','XBtxt',ChangeFileExt(sxbfile,'.xb'));
      EdFileXB.Text    := prj.ReadString('Files','XB',ChangeFileExt(sxbfile,''));
      EdFileXBM.Text   := prj.ReadString('Files','Merge',ChangeFileExt(sxbfile,'.mrg'));
      MaskEdit1.Text   := '100';
      MaskEdit2.Text   := '10';
      TrackBarVerbose.Position := 2;
    end;

    SynEdit1.Lines.SaveToFile(sxbfile);
    xb.Lines.SaveToFile(EdFileXBtxt.Text);

    // writing INI-File
    prj.WriteString('Files','XBtxt',EdFileXBtxt.Text);
    prj.WriteString('Files','XB',EdFileXB.Text);
    prj.WriteString('Files','SXB',EdFileSXB.Text);
    prj.WriteString('Files','Merge',EdFileXBM.Text);
    prj.WriteBool('Build','Token',cbtkn.checked);
    prj.WriteBool('Build','Merge',cbMrg.checked);
    prj.WriteString('Param','LineStart',MaskEdit1.Text);
    prj.WriteString('Param','LineInc',MaskEdit2.Text);
    prj.WriteInteger('Param','Verbose',TrackBarVerbose.Position);
    prj.WriteBool('Param','Debug',cbDebug.checked);
    prj.UpdateFile;
  end;
end;

Procedure UpdateMatrix;
VAR i,m : Integer;
    ax,ay : Integer;
begin
  form1.CharPrev.Canvas.Brush.Color:=clbg;
  form1.CharPrev.Canvas.FillRect(0,0,68,68);
  form1.CharPrev.Canvas.Brush.Color:=clfg;
  ax := 2; ay := 2;
  if form1.RadioGroup1.ItemIndex=0 then m:=63 else m:=255;
  for i:=0 to m do begin
    if CharBit[i].tag=1 then CharBit[i].Brush.Color:=clfg
                        else CharBit[i].Brush.Color:=clbg;
    CharBit[i].Repaint;

    if i=128 then begin ax:=34; ay:=-62 end;
    if CharBit[i].tag=1 then begin
      form1.CharPrev.Canvas.FillRect(ax+4*(i MOD 8),ay+4*(i DIV 8),ax+4*(i MOD 8)+4,ay+4*(i DIV 8)+4);
    end;
  end;
end;

Function nibble2hex(s:String):char;
begin
  case s of
    '0000' : nibble2hex := '0';
    '0001' : nibble2hex := '1';
    '0010' : nibble2hex := '2';
    '0011' : nibble2hex := '3';
    '0100' : nibble2hex := '4';
    '0101' : nibble2hex := '5';
    '0110' : nibble2hex := '6';
    '0111' : nibble2hex := '7';
    '1000' : nibble2hex := '8';
    '1001' : nibble2hex := '9';
    '1010' : nibble2hex := 'A';
    '1011' : nibble2hex := 'B';
    '1100' : nibble2hex := 'C';
    '1101' : nibble2hex := 'D';
    '1110' : nibble2hex := 'E';
    '1111' : nibble2hex := 'F';
     else  nibble2hex := 'X';
  end;
end;

function hex2bin(h:String):AnsiString;
var a : AnsiString;
    i : Integer;
begin
  a := '';
  for i:=1 to length(h) do case UpperCase(h[i]) of
    '0' : a := a + '0000';
    '1' : a := a + '0001';
    '2' : a := a + '0010';
    '3' : a := a + '0011';
    '4' : a := a + '0100';
    '5' : a := a + '0101';
    '6' : a := a + '0110';
    '7' : a := a + '0111';
    '8' : a := a + '1000';
    '9' : a := a + '1001';
    'A' : a := a + '1010';
    'B' : a := a + '1011';
    'C' : a := a + '1100';
    'D' : a := a + '1101';
    'E' : a := a + '1110';
    'F' : a := a + '1111';
    else  a := a + '0000';
  end;
  hex2bin := a;
end;

procedure TForm1.MenuSXBtoXBClick(Sender: TObject);
var xbout, SxbOptions: String;
    LineStart,LineInc :Integer;
begin
  LineStart := StrtoInt(trim(MaskEdit1.Text));
  LineInc   := StrtoInt(trim(MaskEdit2.Text));
  SaveFile;
  xbout := EdFileXBtxt.text;
  if cbDebug.checked then SxbOptions := '-debug'
                     else SxbOptions := '';
  WriteLogLine(-1,'Save '+sxbfile+' and convert to '+xbout+'.');
  WriteLogLine(2,'Start LineNr: '+MaskEdit1.Text+', Increment: '+MaskEdit2.Text+'.');
  WriteLogLine(3,'');
  SXBtoXB(sxbfile,xbout,LineStart,LineInc,@WriteLogLine,SxbOptions);
  xb.Lines.LoadFromFile(xbout);
  xb.ReadOnly:=true;
end;

procedure TForm1.MenuExitClick(Sender: TObject);
begin
  SaveFile;
  Application.Terminate;
end;

procedure TForm1.MenuItem1Click(Sender: TObject);
begin

end;

procedure TForm1.MenuNewClick(Sender: TObject);
begin
  SaveFile;
  sxbfile:='';
  Form1.Caption:='TI Code Editor - New File';
  EdFileSXB.Text   := '';
  EdFilePrj.Text   := '';
  EdFileXBtxt.Text := '';
  EdFileXB.Text    := '';
  EdFileXBM.Text   := '';
  SynEdit1.Lines.Clear;
  xb.Lines.Clear;
  log.Lines.Clear;
  ErrCount:=0; WrnCount := 0;
end;

procedure TForm1.MenuItemLogSaveClick(Sender: TObject);
VAR lfn : String;
begin
  lfn := ChangeFileExt(sxbfile,'.log');
  log.Lines.SaveToFile(lfn);
  ShowMessage('Logfile '+lfn+' written.');
end;

procedure TForm1.MenuItemLogClearClick(Sender: TObject);
begin
  log.Lines.Clear;
  ErrCount:=0; WrnCount := 0;
end;


procedure TForm1.FormCreate(Sender: TObject);
var i,ax,ay: Integer;
begin
  sxbfile := '';
  log.Lines.Clear;
  ErrCount:=0; WrnCount := 0;
  SynEdit1.Clear;
  sxbfile := Paramstr(1);
  if (sxbfile='') or not(FileExists(sxbfile))
     then sxbfile := ''
     else loadfile(sxbfile);

  // Create Char-Editor
  clfg := clBlack; clbg := clWhite;
  ax := 20; ay := 155;
  for i:=0 to 255 do begin
     if i=128 then begin ax := 276; ay:=-357 end;
     CharBit[i]:=TShape.Create(TabSheet5);
     CharBit[i].Shape   := stSquare;
     CharBit[i].height  := 32;
     CharBit[i].Width   := 32;
     CharBit[i].left    := ax + 32 * (i MOD 8);
     CharBit[i].top     := ay + 32 *( i DIV 8);
     CharBit[i].brush.Color  := clBG;
     CharBit[i].OnMouseDown  := @CharPrevMouseDown;
     CharBit[i].OnMouseEnter := @CharPrevMouseEnter;
     CharBit[i].name    := 'Square'+inttostr(i);
     CharBit[i].Parent  := TabSheet5;
     CharBit[i].Visible := (i<64);
  end;
end;

procedure TForm1.ColorBoxFGGetColors(Sender: TCustomColorBox; Items: TStrings);
begin
  with Sender do begin
    // items.Clear;
    AddItem('  2 Black',TObject($000000));
    AddItem('  3 Medium Green',TObject($42c821));
    AddItem('  4 Light Green',TObject($78dc5e));
    AddItem('  5 Blue',TObject($ed5554));
    AddItem('  6 Light Blue',TObject($fc767d));
    AddItem('  7 Dark Red',TObject($4d52d4));
    AddItem('  8 Cyan',TObject($f5eb42));
    AddItem('  9 Red',TObject($5455fc));
    AddItem('10 Light Red',TObject($7879ff));
    AddItem('11 Dark Yellow',TObject($54c1d4));
    AddItem('12 Light Yellow',TObject($80cee6));
    AddItem('13 Dark Green',TObject($3bb021));
    AddItem('14 Magenta',TObject($ba5bc9));
    AddItem('15 Gray',TObject($cccccc));
    AddItem('16 White',TObject($ffffff));
  end;
end;

procedure TForm1.ColorBoxBGChange(Sender: TObject);
begin
  clbg := ColorBoxBG.Selected ;
  UpdateMatrix;
end;

procedure TForm1.CharPrevMouseEnter(Sender: TObject);
var mb : longInt;
begin
  mb := GetKeyState(VK_RBUTTON);
  if mb<0 then CharPrevMouseDown(Sender, mbLeft,[ssLeft],1,1);
end;

procedure TForm1.ColorBoxFGChange(Sender: TObject);
begin
  clfg := ColorBoxFG.Selected;
  UpdateMatrix;
end;

procedure TForm1.FormDropFiles(Sender: TObject; const FileNames: array of String  );
var fn : String;
begin
  fn := FileNames[Low(FileNames)];
  if Application.MessageBox(pChar('Load file '+fn+' ?'), 'Drop File',MB_YESNO+MB_ICONQUESTION) = IDYES then begin
    SaveFile;
    LoadFile(fn);
  end;
end;

procedure TForm1.GroupBox1Resize(Sender: TObject);
begin
  EdFileSXB.Width   := GroupBox1.Width - 450;
  EdFileXBtxt.Width := GroupBox1.Width - 450;
  EdFileXB.Width    := GroupBox1.Width - 450;
  EdFileXBM.Width   := GroupBox1.Width - 450;
  EdFilePrj.Width   := GroupBox1.Width - 450;
end;

procedure TForm1.hexstrChange(Sender: TObject);
var b : AnsiString;
    i : Integer;
begin
  if hexmanmode then begin
     b := hex2bin(hexstr.text);
     for i:=1 to min(length(b),255) do
        if b[i]='1' then CharBit[i-1].Tag:=1
                    else CharBit[i-1].Tag:=0;
     UpdateMatrix;
  end;
end;

procedure TForm1.hexstrEnter(Sender: TObject);
begin
  hexmanmode := true;
end;

procedure TForm1.hexstrExit(Sender: TObject);
var s,c : String;
    i : Integer;
begin
  s := hexstr.text;
  c := '';
  for i:=1 to length(s) do
     if isxdigit(s[i]) and (s[i]<>' ') then c:=c+s[i];
  hexstr.text:=c;
  if c<>s then begin
     hexstr.text:=c;
     hexstrChange(Sender);
  end;
  hexmanmode := false;
end;


procedure TForm1.MenuBuildClick(Sender: TObject);
begin
  log.Lines.Clear;
  ErrCount:=0; WrnCount := 0;
  WriteLogLine(-1,'Build Process started.');
  WriteLogLine(-1,'Step 1: Convert Structured Extended Basic to Standard Extended Basic.');
  WriteLogLine(3,'');
  MenuSXBtoXBClick(Sender);

  WriteLogLine(3,'');
  WriteLogLine(-1,'Step 2: Save Standard Extended Basic in TI internal format.');
  WriteLogLine(3,'');
  MenuTokenizeClick(Sender);

  WriteLogLine(-1,'');
  WriteLogLine(-1,'Errors:   '+IntToStr(ErrCount));
  WriteLogLine(-1,'Warnings: '+IntToStr(WrnCount));
  PageControl1.ActivePage:=TabSheet3;
  if ErrCount+WrnCount>0 then
     Application.MessageBox(pChar('Errors: '+IntToStr(ErrCount)+',  Warnings: '+IntToStr(WrnCount)+'.'), 'Attention',MB_OK+MB_ICONHAND)
end;

procedure TForm1.MenuOpenClick(Sender: TObject);
begin
  if Opendialog1.Execute then begin
    sxbfile := OpenDialog1.FileName;
    if UpperCase(ExtractFileExt(sxbfile)) <> '.SXB' then
      begin
         prj := TiniFile.Create(sxbfile);
         sxbfile := prj.ReadString('Files','SXB',ChangeFileExt(sxbfile,'.SXB'));
         prj.Destroy;
      end;
    LoadFile(sxbfile)
  end;
end;

procedure TForm1.MenuSaveClick(Sender: TObject);
begin
  SaveFile;
end;

procedure TForm1.MenuTIdBitClick(Sender: TObject);
var xbout: String;
    LineStart,LineInc :Integer;
begin
  LineStart := StrtoInt(trim(MaskEdit1.Text));
  LineInc   := StrtoInt(trim(MaskEdit2.Text));
  log.Lines.Clear;
  MenuSaveClick(Sender);
  xbout := EdFileXBtxt.Text;
  WriteLogLine(2,'Save '+sxbfile+' and convert to '+xbout+'.');
  WriteLogLine(2,'Warning: All Debug Messages are Level 2 (Info)!');
  ExportXB(sxbfile,xbout,LineStart,LineInc,@WriteLogLine,'Any Options');
  xb.Lines.LoadFromFile(xbout);
  xb.ReadOnly:=true;
end;

procedure TForm1.MenuTokenizeClick(Sender: TObject);
VAR xbin, tkout,mrgout : String;
begin
  // log.Lines.Clear;
  xbin  := EdFileXBtxt.Text;
  WriteLogLine(2,'Save '+xbin+' and tokenize to ...');
  SaveFile;
  if cbtkn.checked then tkout := EdFileXB.Text
                   else tkout := '';
  if cbMrg.checked then mrgout :=EdFileXBM.Text
                   else mrgout := '';
  if tkout=''  then WriteLogLine(2,'   Token File ommitted.')
               else WriteLogLine(2,'   Token File: '+tkout+'.');
  if mrgout='' then WriteLogLine(2,'   Merge File ommitted.')
               else WriteLogLine(2,'   Merge File: '+mrgout+'.');
  WriteLogLine(2,'');
  tokenize(xbin,tkout,mrgout,@WriteLogLine,'Any Options');
end;

procedure TForm1.MenuUndoClick(Sender: TObject);
begin
  SynEdit1.Undo;
end;

procedure TForm1.MnReplaceClick(Sender: TObject);
begin
   ReplaceDialog1.Execute;
end;

procedure TForm1.PageControl1Change(Sender: TObject);
begin
  if PageControl1.TabIndex=0 then begin
    MenuUndo.Enabled:=true;
  end else begin
    MenuUndo.Enabled:=false;
  end;
end;

procedure TForm1.RadioGroup1Click(Sender: TObject);
var i : Integer;
begin
  if RadioGroup1.ItemIndex=0
                 then for i:=64 to 255 do CharBit[i].Visible:=false
                 else for i:=64 to 255 do CharBit[i].Visible:=true;
  UpdateMatrix;
end;

procedure TForm1.CharPrevMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var i,j,m :integer;
    s,t : String;
    ax,ay : Integer;

begin
  hexmanmode := false;
  with sender as TShape do begin
    if tag=0 then
    begin
      tag:=1;
      brush.color := clfg;
      form1.CharPrev.Canvas.Brush.Color:=clfg;
    end else begin
      tag:=0;
      brush.color := clbg;
      form1.CharPrev.Canvas.Brush.Color:=clbg;
    end;
  end;
  // Update Preview

  m := strtoint(copy(TShape(sender).Name,7,3));
  if m<=128 then begin ax := 2; ay := 2; end
            else begin ax :=34; ay :=-62 end;
  form1.CharPrev.Canvas.FillRect(ax+4*(m MOD 8),ay+4*(m DIV 8),ax+4*(m MOD 8)+4,ay+4*(m DIV 8)+4);

  // Update Preview and Hex-String;
  if RadioGroup1.ItemIndex=0 then m:=7 else m:=31;
  t:='';
  for i:=0 to m do begin
    s := '';
    for j:=0 to 7 do begin
      if CharBit[i*8+j].Tag=1 then s:=s+'1'
                              else s:=s+'0';
    end;
    t := t + nibble2hex(copy(s,1,4)) + nibble2hex(copy(s,5,4));
  end;
  hexstr.Text:=t;
end;

procedure TForm1.ReplaceDialog1Find(Sender: TObject);
var sso: TSynSearchOptions;
    trf: TReplaceFlags;
    i : Integer;
begin
  if PageControl1.TabIndex=0 then with ReplaceDialog1 do begin
    sso := [];
    if frMatchCase   IN Options then include(sso,ssoMatchCase);
    if frReplace     IN Options then include(sso,ssoReplace);
    if frReplaceAll  IN Options then include(sso,ssoReplaceAll);
    if frWholeWord   IN Options then include(sso,ssoWholeWord);
    if frEntireScope IN Options then include(sso,ssoEntireScope);
    SynEdit1.SearchReplace(ReplaceDialog1.FindText,ReplaceDialog1.ReplaceText,sso);
  end else if PageControl1.TabIndex=1 then with ReplaceDialog1 do begin
    trf := [];
    if not (frMatchCase IN Options) then include(trf,rfIgnoreCase);
    if  frReplaceAll    IN Options  then include(trf,rfReplaceAll);
    for i:=0 to xb.Lines.Count-1 do begin
      xb.Lines.Strings[i] := StringReplace(xb.Lines.Strings[i],ReplaceDialog1.FindText,ReplaceDialog1.ReplaceText,trf);
    end;
  end;
end;

procedure TForm1.ReplaceDialog1Replace(Sender: TObject);
var sso: TSynSearchOptions;
    trf: TReplaceFlags;
    i : Integer;
begin

  if PageControl1.TabIndex=0 then with ReplaceDialog1 do begin
    sso := [];
    if frMatchCase   IN Options then include(sso,ssoMatchCase);
    if frReplace     IN Options then include(sso,ssoReplace);
    if frReplaceAll  IN Options then include(sso,ssoReplaceAll);
    if frWholeWord   IN Options then include(sso,ssoWholeWord);
    if frEntireScope IN Options then include(sso,ssoEntireScope);

    SynEdit1.SearchReplace(ReplaceDialog1.FindText,ReplaceDialog1.ReplaceText,sso);
  end else if PageControl1.TabIndex=1 then with ReplaceDialog1 do begin
    trf := [];
    if not (frMatchCase IN Options) then include(trf,rfIgnoreCase);
    if  frReplaceAll    IN Options  then include(trf,rfReplaceAll);
    for i:=0 to xb.Lines.Count-1 do begin
      xb.Lines.Strings[i] := StringReplace(xb.Lines.Strings[i],ReplaceDialog1.FindText,ReplaceDialog1.ReplaceText,trf);
    end;
  end;
end;

//TSynSearchOption =
//    ( ssoMatchCase, ssoWholeWord,
//      ssoBackwards,
//      ssoEntireScope, ssoSelectedOnly,
//      ssoReplace, ssoReplaceAll,
//      ssoPrompt,
//      ssoSearchInReplacement,    // continue search-replace in replacement (with ssoReplaceAll) // replace recursive
//      ssoRegExpr, ssoRegExprMultiLine,
//      ssoFindContinue      // Assume the current selection is the last match, and start search behind selection
//                           // (before if ssoBackward) // Default is to start at caret (Only SearchReplace / SearchReplaceEx has start/end param)
//    );
//  TSynSearchOptions = set of TSynSearchOption;

procedure TForm1.ReplaceDialog1Show(Sender: TObject);
var fo :  TFindOptions;
begin
  case PageControl1.TabIndex of
    0 : fo := [frDisableUpDown,frHidePromptOnReplace];
    1 : fo := [frDisableUpDown,frDisableWholeWord,frHidePromptOnReplace];
  end;
  ReplaceDialog1.Options := fo;
end;


procedure TForm1.TrackBarVerboseChange(Sender: TObject);
begin
  case TrackBarVerbose.Position of
    0 : TrackBarVerbose.Hint:='Error';
    1 : TrackBarVerbose.Hint:='Warning';
    2 : TrackBarVerbose.Hint:='Information';
    3 : TrackBarVerbose.Hint:='Verbose';
    4 : TrackBarVerbose.Hint:='Very Verbose';
    5 : TrackBarVerbose.Hint:='Unbelievable Verbose';
  end;
end;


procedure TForm1.xbDblClick(Sender: TObject);
begin
  xb.ReadOnly:=false;
end;

end.

